@@ -1,4 +1,4 @@
-# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
+# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
 # SPDX-License-Identifier: MIT
 #
 #
@@ -26,4 +26,7 @@ try:
 except ModuleNotFoundError:
     import importlib_metadata
 
-__version__ = importlib_metadata.version(__name__.replace(".", "-"))
+try:
+    __version__ = importlib_metadata.version(__name__.replace(".", "-"))
+except importlib_metadata.PackageNotFoundError:
+    __version__ = "DEBUG"
