@@ -1,4 +1,4 @@
-# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
+# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
 # SPDX-License-Identifier: MIT
 #
 #
@@ -21,19 +21,31 @@
 # SOFTWARE.
 
 """Simple API client for the Ansys ConceptEV service."""
-
+from collections import defaultdict
 import datetime
+import json
 from json import JSONDecodeError
-import os
+import re
 from typing import Literal
 
-import dotenv
 import httpx
+from tenacity import retry, retry_if_result, stop_after_delay, wait_random_exponential
 
 from ansys.conceptev.core import auth
+from ansys.conceptev.core.exceptions import (
+    AccountsError,
+    DeleteError,
+    DesignError,
+    ProductAccessError,
+    ProductIdsError,
+    ProjectError,
+    ResponseError,
+    ResultsError,
+    TokenError,
+    UserDetailsError,
+)
 from ansys.conceptev.core.progress import check_status, monitor_job_progress
-
-dotenv.load_dotenv()
+from ansys.conceptev.core.settings import settings
 
 Router = Literal[
     "/architectures",
@@ -66,34 +78,41 @@ PRODUCT_ACCESS_ROUTES = [
     "/jobs:start",
 ]
 
-JOB_TIMEOUT = auth.config["JOB_TIMEOUT"]
+JOB_TIMEOUT = settings.job_timeout
+OCM_URL = settings.ocm_url
+BASE_URL = settings.conceptev_url
+ACCOUNT_NAME = settings.account_name
 app = auth.create_msal_app()
 
 
-def get_token() -> str:
-    """Get token from OCM."""
-    username = os.environ["CONCEPTEV_USERNAME"]
-    password = os.environ["CONCEPTEV_PASSWORD"]
-    ocm_url = auth.config["OCM_URL"]
-    response = httpx.post(
-        url=ocm_url + "/auth/login/", json={"emailAddress": username, "password": password}
-    )
-    if response.status_code != 200:
-        raise Exception(f"Failed to get token {response.content}")
-    return response.json()["accessToken"]
+def is_gaetway_error(response) -> bool:
+    """Check if the response is a gateway error."""
+    if isinstance(response, httpx.Response):
+        return response.status_code in (502, 504)
+    return False
 
 
-def get_http_client(token: str, design_instance_id: str | None = None) -> httpx.Client:
+def get_http_client(
+    token: str | None = None,
+    design_instance_id: str | None = None,
+    cache_filepath: str = "token_cache.bin",
+) -> httpx.Client:
     """Get an HTTP client.
 
     The HTTP client creates and maintains the connection, which is more performant than
     re-creating this connection for each call.
     """
-    base_url = auth.config["CONCEPTEV_URL"]
-    params = None
-    if design_instance_id:
-        params = {"design_instance_id": design_instance_id}
-    return httpx.Client(headers={"Authorization": token}, params=params, base_url=base_url)
+    httpx_auth = auth.AnsysIDAuth(cache_filepath=cache_filepath) if token is None else None
+    params = {"design_instance_id": design_instance_id} if design_instance_id else None
+    header = {"Authorization": token} if token else None
+
+    client = httpx.Client(headers=header, auth=httpx_auth, params=params, base_url=BASE_URL)
+    client.send = retry(
+        retry=retry_if_result(is_gaetway_error),
+        wait=wait_random_exponential(multiplier=1, max=60),
+        stop=stop_after_delay(10),
+    )(client.send)
+    return client
 
 
 def process_response(response) -> dict:
@@ -106,7 +125,7 @@ def process_response(response) -> dict:
             return response.json()
         except JSONDecodeError:
             return response.content
-    raise Exception(f"Response Failed:{response.content}")
+    raise ResponseError(f"Response Failed:{response.content}")
 
 
 def get(
@@ -145,7 +164,7 @@ def check_product_access(router: Router, account_id: str | None, params: dict) -
     """Check account_id is there for product access."""
     if router in PRODUCT_ACCESS_ROUTES:
         if not account_id:
-            raise Exception(f"Account ID is required for {router}.")
+            raise ProductAccessError(f"Account ID is required for {router}.")
         params = params | {"account_id": account_id}
     return params
 
@@ -159,7 +178,7 @@ def delete(client: httpx.Client, router: Router, id: str, account_id: str | None
     path = "/".join([router, id])
     response = client.delete(url=path, params=params)
     if response.status_code != 204:
-        raise Exception(f"Failed to delete from {router} with ID:{id}.")
+        raise DeleteError(f"Failed to delete from {router} with ID:{id}.")
 
 
 def put(client: httpx.Client, router: Router, id: str, data: dict) -> dict:
@@ -180,8 +199,7 @@ def create_new_project(
     project_goal: str = "Created from the CLI",
 ) -> dict:
     """Create a project."""
-    osm_url = auth.config["OCM_URL"]
-    token = client.headers["Authorization"]
+    token = get_token(client)
     project_data = {
         "accountId": account_id,
         "hpcId": hpc_id,
@@ -189,23 +207,45 @@ def create_new_project(
         "projectGoal": project_goal,
     }
     created_project = httpx.post(
-        osm_url + "/project/create", headers={"Authorization": token}, json=project_data
+        OCM_URL + "/project/create", headers={"Authorization": token}, json=project_data
     )
     if created_project.status_code != 200 and created_project.status_code != 204:
-        raise Exception(f"Failed to create a project {created_project}.")
+        raise ProjectError(f"Failed to create a project {created_project}.")
 
     return created_project.json()
 
 
+def get_or_create_project(client: httpx.Client, account_id: str, hpc_id: str, title: str) -> dict:
+    """Get or create a project."""
+    stored_errors = []
+    options = [title, re.escape(title), title.split(maxsplit=1)[0]]
+    for search_string in options:
+        try:
+            projects = get_project_ids(search_string, account_id, client.headers["Authorization"])
+            project_id = projects[title][0]
+            return project_id
+        except (ProjectError, KeyError, IndexError) as err:
+            stored_errors.append(err)
+
+    project = create_new_project(client, account_id, hpc_id, title)
+    project_id = project["projectId"]
+
+    return project_id
+
+
 def create_new_concept(
     client: httpx.Client,
     project_id: str,
-    title: str = f"CLI concept {datetime.datetime.now()}",
+    product_id: str | None = None,
+    title: str | None = None,
 ) -> dict:
     """Create a concept within an existing project."""
-    osm_url = auth.config["OCM_URL"]
-    token = client.headers["Authorization"]
-    product_id = get_product_id(token)
+    if title is None:
+        title = f"CLI concept {datetime.datetime.now()}"
+
+    token = get_token(client)
+    if product_id is None:
+        product_id = get_product_id(token)
 
     design_data = {
         "projectId": project_id,
@@ -213,11 +253,11 @@ def create_new_concept(
         "designTitle": title,
     }
     created_design = httpx.post(
-        osm_url + "/design/create", headers={"Authorization": token}, json=design_data
+        OCM_URL + "/design/create", headers={"Authorization": token}, json=design_data
     )
 
     if created_design.status_code not in (200, 204):
-        raise Exception(f"Failed to create a design on OCM {created_design.content}.")
+        raise DesignError(f"Failed to create a design on OCM {created_design.content}.")
 
     user_id = get_user_id(token)
 
@@ -246,10 +286,9 @@ def create_new_concept(
 
 def get_product_id(token: str) -> str:
     """Get the product ID."""
-    osm_url = auth.config["OCM_URL"]
-    products = httpx.get(osm_url + "/product/list", headers={"Authorization": token})
+    products = httpx.get(OCM_URL + "/product/list", headers={"Authorization": token})
     if products.status_code != 200:
-        raise Exception(f"Failed to get product id.")
+        raise ProductIdsError(f"Failed to get product id.")
 
     product_id = [
         product["productId"] for product in products.json() if product["productName"] == "CONCEPTEV"
@@ -259,10 +298,9 @@ def get_product_id(token: str) -> str:
 
 def get_user_id(token):
     """Get the user ID."""
-    osm_url = auth.config["OCM_URL"]
-    user_details = httpx.post(osm_url + "/user/details", headers={"Authorization": token})
+    user_details = httpx.post(OCM_URL + "/user/details", headers={"Authorization": token})
     if user_details.status_code not in (200, 204):
-        raise Exception(f"Failed to get a user details on OCM {user_details}.")
+        raise UserDetailsError(f"Failed to get a user details on OCM {user_details}.")
     user_id = user_details.json()["userId"]
     return user_id
 
@@ -275,10 +313,9 @@ def get_concept_ids(client: httpx.Client) -> dict:
 
 def get_account_ids(token: str) -> dict:
     """Get account IDs."""
-    ocm_url = auth.config["OCM_URL"]
-    response = httpx.post(url=ocm_url + "/account/list", headers={"authorization": token})
+    response = httpx.post(url=OCM_URL + "/account/list", headers={"authorization": token})
     if response.status_code != 200:
-        raise Exception(f"Failed to get accounts {response}.")
+        raise AccountsError(f"Failed to get accounts {response}.")
     accounts = {
         account["account"]["accountName"]: account["account"]["accountId"]
         for account in response.json()
@@ -286,16 +323,22 @@ def get_account_ids(token: str) -> dict:
     return accounts
 
 
+def get_account_id(token: str) -> str:
+    """Get the account ID from OCM using name from config file."""
+    accounts = get_account_ids(token)
+    account_id = accounts[ACCOUNT_NAME]
+    return account_id
+
+
 def get_default_hpc(token: str, account_id: str) -> dict:
     """Get the default HPC ID."""
-    ocm_url = auth.config["OCM_URL"]
     response = httpx.post(
-        url=ocm_url + "/account/hpc/default",
+        url=OCM_URL + "/account/hpc/default",
         json={"accountId": account_id},
         headers={"authorization": token},
     )
     if response.status_code != 200:
-        raise Exception(f"Failed to get accounts {response}.")
+        raise AccountsError(f"Failed to get accounts {response}.")
     return response.json()["hpcId"]
 
 
@@ -304,9 +347,11 @@ def create_submit_job(
     concept: dict,
     account_id: str,
     hpc_id: str,
-    job_name: str = "cli_job: " + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"),
+    job_name: str | None = None,
 ) -> dict:
     """Create and then submit a job."""
+    if job_name is None:
+        job_name = f"cli_job: {datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')}"
     job_input = {
         "job_name": job_name,
         "requirement_ids": concept["requirements_ids"],
@@ -327,7 +372,7 @@ def create_submit_job(
 
 def read_file(filename: str) -> str:
     """Read a given file."""
-    with open(filename) as f:
+    with open(filename, "r+b") as f:
         content = f.read()
     return content
 
@@ -337,38 +382,115 @@ def read_results(
     job_info: dict,
     calculate_units: bool = True,
     timeout: int = JOB_TIMEOUT,
+    filtered: bool = False,
+    msal_app: auth.PublicClientApplication | None = None,
 ) -> dict:
     """Read job results."""
     job_id = job_info["job_id"]
-    token = client.headers["Authorization"]
+    token = get_token(client)
     user_id = get_user_id(token)
     initial_status = get_status(job_info, token)
     if check_status(initial_status):  # Job already completed
-        return get_results(client, job_info, calculate_units)
+        return get_results(client, job_info, calculate_units, filtered)
     else:  # Job is still running
-        monitor_job_progress(job_id, user_id, token, timeout)  # Wait for completion
-        return get_results(client, job_info, calculate_units)
+        if msal_app is None:
+            msal_app = auth.create_msal_app()
+        monitor_job_progress(job_id, user_id, token, msal_app, timeout)  # Wait for completion
+
+        token = auth.get_ansyId_token(msal_app)
+        client.headers["Authorization"] = token  # Update the token
+        return get_results(client, job_info, calculate_units, filtered)
 
 
-def get_results(client, job_info: dict, calculate_units: bool = True):
+def get_results(
+    client,
+    job_info: dict,
+    calculate_units: bool = True,
+    filtered: bool = False,
+):
     """Get the results."""
     version_number = get(client, "/utilities:data_format_version")
+    if filtered:
+        filename = f"filtered_output_v{version_number}.json"
+    else:
+        filename = f"output_file_v{version_number}.json"
     response = client.post(
         url="/jobs:result",
         json=job_info,
         params={
-            "results_file_name": f"output_file_v{version_number}.json",
+            "results_file_name": filename,
             "calculate_units": calculate_units,
         },
     )
+    if response.status_code == 502 or response.status_code == 504:
+        raise ResultsError(
+            f"Request timed out {response}. "
+            f"Please try using either calculate_units=False or filtered=True."
+        )
     return process_response(response)
 
 
+def get_job_file(token, job_id, filename, simulation_id=None, encrypted=False):
+    """Get the job file from the OnScale Cloud Manager."""
+    encrypted_part = "decrypted/" if encrypted else ""
+    if simulation_id is not None:
+        path = f"{OCM_URL}/job/files/{encrypted_part}{job_id}/{simulation_id}/{filename}"
+    else:
+        path = f"{OCM_URL}/job/files/{encrypted_part}{job_id}/{filename}"
+    response = httpx.get(
+        url=path, headers={"authorization": token, "accept": "application/octet-stream"}
+    )
+    if response.status_code != 200:
+        raise ResponseError(f"Failed to get file {response}.")
+
+    return json.loads(response.content)
+
+
+def get_job_info(token, job_id):
+    """Get the job info from the OnScale Cloud Manager."""
+    response = httpx.post(
+        url=f"{OCM_URL}/job/load", headers={"authorization": token}, json={"jobId": job_id}
+    )
+    response = process_response(response)
+    job_info = {
+        "job_id": job_id,
+        "simulation_id": response["simulations"][0]["simulationId"],
+        "job_name": response["jobName"],
+        "docker_tag": response["dockerTag"],
+    }
+    return job_info
+
+
+def get_design_of_job(token, job_id):
+    """Get the job info from the OnScale Cloud Manager."""
+    response = httpx.post(
+        url=f"{OCM_URL}/job/load", headers={"authorization": token}, json={"jobId": job_id}
+    )
+    response = process_response(response)
+    return response["designInstanceId"]
+
+
+def get_design_title(token, design_instance_id):
+    """Get the design Title from the OnScale Cloud Manager."""
+    response = httpx.post(
+        url=f"{OCM_URL}/design/instance/load",
+        headers={"authorization": token},
+        json={"designInstanceId": design_instance_id},
+    )
+    response = process_response(response)
+    design = httpx.post(
+        url=f"{OCM_URL}/design/load",
+        headers={"authorization": token},
+        json={"designId": response["designId"]},
+    )
+    design = process_response(design)
+    return design["designTitle"]
+
+
 def get_status(job_info: dict, token: str) -> str:
     """Get the status of the job."""
-    ocm_url = auth.config["OCM_URL"]
     response = httpx.post(
-        url=ocm_url + "/job/load",
+        url=OCM_URL + "/job/load",
         json={"jobId": job_info["job_id"]},
         headers={"Authorization": token},
     )
@@ -377,6 +499,61 @@ def get_status(job_info: dict, token: str) -> str:
     return initial_status
 
 
+def get_project_ids(name: str, account_id: str, token: str) -> dict:
+    """Get projects."""
+    response = httpx.post(
+        url=OCM_URL + "/project/list/page",
+        json={"accountId": account_id, "filterByName": name, "pageNumber": 0, "pageSize": 1000},
+        headers={"Authorization": token},
+    )
+    processed_response = process_response(response)
+    projects = processed_response["projects"]
+    project_dict = defaultdict(list)
+    for project in projects:
+        project_dict[project["projectTitle"]].append(project["projectId"])
+    return project_dict
+
+
+def get_project_id(name: str, account_id: str, token: str) -> str:
+    """Get project ID."""
+    projects = get_project_ids(name, account_id, token)
+    if not projects:
+        raise ProjectError(f"Project with name {name} not found.")
+    if len(projects) > 1:
+        raise ProjectError(f"Multiple projects found with name {name}.")
+    return projects[name][0]
+
+
+def get_token(client: httpx.Client) -> str:
+    """Get the token from the client."""
+    if client.auth is not None and client.auth.app is not None:
+        return auth.get_ansyId_token(client.auth.app)
+    elif client.headers is not None and "Authorization" in client.headers:
+        return client.headers["Authorization"]
+    raise TokenError("App not found in client.")
+
+
+def delete_project(project_id, token):
+    """Delete a project."""
+    ocm_delete_init = httpx.request(
+        method="DELETE",
+        url=OCM_URL + "/project/delete/init",
+        headers={"Authorization": token},
+        json={"projectId": project_id},
+        timeout=20,
+    )
+    ocm_delete_init = process_response(ocm_delete_init)
+    ocm_delete = httpx.request(
+        method="DELETE",
+        url=OCM_URL + "/project/delete/execute",
+        headers={"Authorization": token},
+        json={"projectId": project_id, "hash": ocm_delete_init["hash"]},
+        timeout=20,
+    )
+    ocm_delete = process_response(ocm_delete)
+    return ocm_delete
+
+
 def post_component_file(client: httpx.Client, filename: str, component_file_type: str) -> dict:
     """Send a POST request to the base client with a file.
 
@@ -405,9 +582,45 @@ def get_concept(client: httpx.Client, design_instance_id: str) -> dict:
     return concept
 
 
-if __name__ == "__main__":
-    token = get_token()
+def create_design_instance(project_id, title, token, product_id=None):
+    """Create a design instance on OCM."""
+    if product_id is None:
+        product_id = get_product_id(token)
 
-    with get_http_client(token) as client:  # Create a client to talk to the API
-        health = get(client, "/health")  # Check that the API is healthy
-        print(health)
+    design_data = {
+        "projectId": project_id,
+        "productId": product_id,
+        "designTitle": title,
+    }
+    created_design = httpx.post(
+        OCM_URL + "/design/create", headers={"Authorization": token}, json=design_data
+    )
+
+    if created_design.status_code not in (200, 204):
+        raise Exception(f"Failed to create a design on OCM {created_design.content}.")
+
+    design_instance_id = created_design.json()["designInstanceList"][0]["designInstanceId"]
+    return design_instance_id
+
+
+def copy_concept(base_concept_id, design_instance_id, client):
+    """Copy the reference concept to the new design instance."""
+    copy = {
+        "old_design_instance_id": base_concept_id,
+        "new_design_instance_id": design_instance_id,
+        "copy_jobs": False,
+    }
+    # Clone the base concept
+    params = {"design_instance_id": design_instance_id, "populated": False}
+    client.params = params
+    concept = post(client, "/concepts:copy", data=copy)
+    return concept
+
+
+def get_component_id_map(client, design_instance_id):
+    """Get a map of component name to component id."""
+    ###TODO move to results file so its self contained.
+    components = client.get(f"/concepts/{design_instance_id}/components")
+    components = process_response(components)
+    components.append({"name": "N/A", "id": None})
+    return {component["name"]: component["id"] for component in components}
